"""
Payton math library

Payton is designed with the minimum requirements. So, to reduce the number
of requirements, instead of using 3rd party mathematical libraries, I created
my simple functions here.

This math module has no requirements, so it can be used separately at other
projects.
"""
